#include "types.h"
#include "a32.h"
#include "pm.h"
#include "vesa.h"
#include "isr_wrap.h"

#define IDT_entries     0x20

#define ADDR_LFB        0x200000
#define ENTRY_LFB       (ADDR_LFB >> 12)
#define XTRA_PAGES      3

typedef struct {
  dword phys_addr;      // phys addr of a page
  byte  *data;          // linear addr to access data within the page
  dword lfb;            // cr2 contents upon page fault - ADDR_LFB
} XTRA;

byte *scr;

byte buf[PAGE_SIZE*(2+XTRA_PAGES+1)]; // 2pg - PD/PT, XTRA RAM pg, 1 alignment pg
byte stack0[PAGE_SIZE];
TSS  tss;

dword phys_addr, addr, *page_dir, *page_tab;
byte *vid_buf;
dword width, height, size, gran_per_bank, cur_bank, tot_banks;

XTRA pages_list[XTRA_PAGES];
dword pages_list_cnt;   // number of items in the list in use
dword started;

INT_DESCRIPTOR  idt[IDT_entries];
IDT_REG         idtr;

void select_bank (dword bank_no) {
  move_window (bank_no * gran_per_bank);
}

void gpf_handler (dword error_code, dword eip, dword cs) {
  beep();
  for(;;);
  nobeep();
}

void setup_IDT() {
  memset (&idt, 0, sizeof(idt));

// Int 01h - Single-Step Exception Handler

  idt[0x01].offs0_15 = ((dword)(&isr_01_wrapper))&0xFFFF;
  idt[0x01].offs16_31 = ((dword)(&isr_01_wrapper)) >> 16;
  idt[0x01].sel = 8;
  idt[0x01].paramcnt = 0;
  idt[0x01].attrs = ACS_INT_GATE;

// Int 0Dh - General Protection Fault Handler

  idt[0x0D].offs0_15 = ((dword)(&isr_0D_wrapper))&0xFFFF;
  idt[0x0D].offs16_31 = ((dword)(&isr_0D_wrapper)) >> 16;
  idt[0x0D].sel = 8;
  idt[0x0D].paramcnt = 0;
  idt[0x0D].attrs = ACS_INT_GATE;

// Int 0Eh - Page Fault Handler

  idt[0x0E].offs0_15 = ((dword)(&isr_0E_wrapper))&0xFFFF;
  idt[0x0E].offs16_31 = ((dword)(&isr_0E_wrapper)) >> 16;
  idt[0x0E].sel = 8;
  idt[0x0E].paramcnt = 0;
  idt[0x0E].attrs = ACS_INT_GATE;

  idtr.base = (dword)base + (dword) &idt;
  idtr.limit = sizeof(idt)-1;
  LIDT (&idtr);
}

void restore_IDT() {
  idtr.base = 0;
  idtr.limit = 1023;
  LIDT (&idtr);
}

void setup_TSS() {
  SEG_DESCRIPTOR *gdt;
  GDT_REG gdtr;
  dword addr_tss;

  SGDT (&gdtr);
  gdt = (SEG_DESCRIPTOR *) (gdtr.base - (dword)base);
  addr_tss = (dword)base + (dword)&tss;
  gdt[6].base0_15 = addr_tss;
  gdt[6].base16_23 = addr_tss >> 16;

  memset (&tss, 0, sizeof(tss));
  tss.esp0 = (dword)&stack0[PAGE_SIZE];
  tss.ss0 = 0x10;
  tss.io_map_addr = 0x68;

  LTR (0x30);
}

void pages_init() {
  dword a, i, j, cnt, *p;

  phys_addr = (dword)base + (dword)&buf;
  if (phys_addr & (PAGE_SIZE-1)) phys_addr = (phys_addr + PAGE_SIZE) & PAGE_MASK;
  addr = phys_addr - (dword)base;

  page_dir = (dword *) addr;
  page_tab = page_dir + PAGE_TABLE_ENTRIES;

  vid_buf = (byte *) (0xA0000 - (dword)base);

  memset (page_dir, 0, PAGE_SIZE);
  memset (page_tab, 0, PAGE_SIZE);

  page_dir[0] = (phys_addr + PAGE_SIZE) | P_PRESENT | P_READWRITE | P_USERSUPER;
  a = 0;
  for(i=0;i<PAGE_TABLE_ENTRIES;i++) {
    page_tab[i] = a | P_PRESENT | P_READWRITE | P_USERSUPER;
    a += PAGE_SIZE;
  }

  i = ENTRY_LFB;
  cnt = tot_banks;
  while (cnt--) {
    a = 0xA0000;
    for(j=0;j<16;j++) {
      page_tab[i++] = a | P_READWRITE | P_USERSUPER;
      a += PAGE_SIZE;
    }
  }

  LCR3 (phys_addr);

  p = page_tab + PAGE_TABLE_ENTRIES;
  a = (dword)p + (dword)base;
  for(i=0;i<XTRA_PAGES;i++) {
    pages_list[i].phys_addr = a;
    pages_list[i].data = (byte *)p;
    a += PAGE_SIZE;
    p += PAGE_TABLE_ENTRIES;
  }
  started = pages_list_cnt = 0;
}

void map_bank (dword bank_no) {
  dword i, j=16;
  i = ENTRY_LFB + (bank_no<<4);
  while (j--)
    page_tab[i++] |= P_PRESENT;
  LCR3(SCR3());
}

void unmap_bank (dword bank_no) {
  dword i, j=16;
  i = ENTRY_LFB + (bank_no<<4);
  while (j--)
    page_tab[i++] &= (PAGE_MASK | P_READWRITE | P_USERSUPER);
  LCR3(SCR3());
}

void scr_init() {
  scr = (byte *) (ADDR_LFB - (dword)base);
  width = mode_info->Width;
  height = mode_info->Height;
  size = width * height;
  tot_banks = size / 0x10000;
  if (size % 0x10000) tot_banks++;
  cur_bank = 0;
  gran_per_bank = 64 / mode_info->WinGranul;
}

int page_fault_handler (dword error_code) {
  dword linear, bank, pg, pgb;

  linear = SCR2();
  if ((linear < ADDR_LFB) || (linear >= ADDR_LFB + size))
    for(;;);

  linear -= ADDR_LFB;
  pg = linear >> 12;
  pgb = pg & 0xF;
  bank = pg >> 4;

  if (started==0) {
    started++;

    unmap_bank (cur_bank);
    map_bank (cur_bank = bank);
    select_bank (cur_bank);

    return 1;
  }

  started++;

  select_bank (bank);
  memcpy ((void*)pages_list[pages_list_cnt].data,
          (void*)&vid_buf[pgb<<12], PAGE_SIZE);
  select_bank (cur_bank);

  page_tab[ENTRY_LFB + pg] = pages_list[pages_list_cnt].phys_addr
                             | P_PRESENT | P_READWRITE | P_USERSUPER;
  LCR3(SCR3());

  pages_list[pages_list_cnt].lfb = linear;
  pages_list_cnt++;

  return 1;
}

void single_step_handler() {
  dword i, bank, pg, pgb;

  started = 0;
  if (!pages_list_cnt) return;

  for (i=0;i<pages_list_cnt;i++) {
    pg = pages_list[i].lfb >> 12;
    pgb = pg & 0xF;
    bank = pg >> 4;

    select_bank (bank);
    memcpy ((void*)&vid_buf[pgb<<12],
            (void*)pages_list[i].data, PAGE_SIZE);

    page_tab[ENTRY_LFB + pg] = (0xA0000 + (pgb<<12))
                               | P_READWRITE | P_USERSUPER;
  }

  LCR3(SCR3());
  pages_list_cnt = 0;
  select_bank (cur_bank);
}

void show_stuff() {
  int x, y, w, h, m, n, i, j, k, l;

/*

// a thorough test
// (MOVS from one bank boundary crossing location to another one)

  scr[65536+65534] = YELLOW;
  scr[65536+65535] = LIGHTBLUE;
  scr[65536+65536] = LIGHTRED;
  scr[65536+65537] = LIGHTGREEN;
  scr[0] = WHITE;
  memcpy ((void*)&scr[65536+65536+65534], (void*)&scr[65536+65534], 4);

*/

// a simple test 
// (no bank crossings, no simultaneously access to 2 or more pages)

  memset (scr, WHITE, size/3);
  memset (scr+size/3, BLUE, size/3);
  memset (scr+2*size/3, RED, size/3);

  w = img[1]; w<<=8; w+=img[0];
  h = img[3]; h<<=8; h+=img[2];

  m = width / w - 1;
  n = height / h - 1;

  y = (height-h*n)/2;
  l = n;
  while (l--) {
    x = (width-w*m)/2;
    k = m;
    while (k--) {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++)
          if (img[4+j*w+i] != 0xff)
            scr[x+i+(y+j)*width] = img[4+j*w+i];
      x += w;
    }
    y += h;
  }
}

int main() {
  scr_init();
  pages_init();
  LCR0(SCR0() | 0x80000000);
  setup_IDT();
  setup_TSS();

  show_stuff();

  LCR0(SCR0() & 0x7FFFFFFF);
  LCR3(0);
  restore_IDT();
  return 0;
}
